module PE where

import Term
import Substitution
import Data.List

type PE = TermPair -- primitive equation

showPE :: PE -> String
showPE (s, t) = show s ++ " = " ++ show t

-- implement variant as rules
-- TODO: is this correct?
-- l1 -> r1 and l2 -> r2 are the same rules
-- if (l1 -> r1) \preceq (l2 -> r2) and (l2 -> r2) \preceq (l1 -> r1)
-- where (l1 -> r1) \preceq (l2 -> r2)
--       if there exists substitution σ such that l1 σ = l2 and r1 σ = r2
variant' :: PE -> PE -> Bool
variant' rule1 rule2 = compare' rule1 rule2 && compare' rule2 rule1
  where compare' (l1, r1) (l2, r2) =
          case match l1 l2 of -- NOTE: we can't use unify. consider f(x,x) -> x and f(x,y) -> x.
            Nothing -> False
            Just sigma -> substitute r1 sigma == r2

-- NOTE: Equation.variant (l, r) (r,l) = True
variant :: PE -> PE -> Bool
variant (s1, t1) (s2, t2) =
  variant' (s1, t1) (s2, t2) || variant' (s1, t1) (t2, s2) ||
  variant' (t1, s1) (s2, t2) || variant' (t1, s1) (t2, s2) 

rename :: PE -> Int -> (Int, PE)
rename (l,r) n = (n + length vars, (substitute l sigma, substitute r sigma))
                 where vars = nub (variables l ++ variables r)
                       sigma = zip vars [ V (v ++ show i) | (v, i) <- zip vars [n..] ]

reverse :: PE -> PE
reverse (l, r) = (r, l)

size :: PE -> Int
size (s, t) = Term.size s + Term.size t

isGround :: PE -> Bool
isGround (s, t) = Term.isGround s && Term.isGround t
